using Benday.EfCore.Api;
using Microsoft.EntityFrameworkCore;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using System.Threading;

namespace Benday.EfCore.Tests
{
    [TestClass]
    public class PersonFixture
    {
        [TestMethod]
        public void CreatePersonUpdatesLastUpdated()
        {
            var person = new Person();

            person.FirstName = UnitTestUtility.GetStringValue("fn");
            person.LastName = UnitTestUtility.GetStringValue("ln");

            var originalLastUpdated = person.LastUpdated;

            using (MyDbContext context = new MyDbContext())
            {
                context.Persons.Add(person);
                context.SaveChanges();
            }

            var actual = person.LastUpdated;

            Assert.AreNotEqual<DateTime>(originalLastUpdated, actual, "LastUpdated didn't change.");
        }

        [TestMethod]
        public void LastUpdatedValueChangesWhenModifiedValuesAreSaved()
        {
            Person person = CreatePersonAndSave();

            Person reloaded = null;
            DateTime originalLastUpdated;

            using (MyDbContext context = new MyDbContext())
            {
                reloaded = context.Persons.Find(person.Id);

                if (reloaded == null)
                {
                    Assert.Fail("Could not find person with id '{0}'.", person.Id);
                }

                originalLastUpdated = person.LastUpdated;

                reloaded.FirstName = UnitTestUtility.GetStringValue("reloadedfn");
                reloaded.LastName = UnitTestUtility.GetStringValue("reloadedln");

                context.SaveChanges();
            }

            var actual = reloaded.LastUpdated;

            Assert.AreNotEqual<DateTime>(originalLastUpdated, actual, "LastUpdated didn't change.");
        }

        [TestMethod]
        public void CreatePersonAndModifyOnDifferentDbContext()
        {
            var person = CreatePersonAndSave();

            var originalLastUpdated = person.LastUpdated;

            person.FirstName = UnitTestUtility.GetStringValue("modifiedfn");
            person.LastName = UnitTestUtility.GetStringValue("modifiedln");

            UpdatePerson(person);

            var reloaded = LoadExistingPerson(person.Id);

            Assert.AreEqual<string>(person.FirstName, reloaded.FirstName, "FirstName");
            Assert.AreEqual<string>(person.LastName, reloaded.LastName, "LastName");

            Assert.AreNotEqual<DateTime>(
                originalLastUpdated,
                reloaded.LastUpdated,
                "Unexpected last updated value.");

            Assert.AreEqual<DateTime>(person.LastUpdated, reloaded.LastUpdated,
                "Last update value should match.");
        }

        [TestMethod]
        [ExpectedException(typeof(DbUpdateConcurrencyException))]
        public void ModifyPersonFailsOnConcurrencyProblem()
        {
            var person = CreatePersonAndSave();

            LoadPersonModifyAndSave(person.Id);

            UpdatePerson(person);
        }

        [TestMethod]
        public void CreatePersonWithPhoneNumbers()
        {
            Person person = CreatePersonWithPhoneNumbersAndSave();

            Assert.AreNotEqual<int>(0, person.PhoneNumbers[0].Id, "Phone 0 wasn't saved.");
            Assert.AreNotEqual<int>(1, person.PhoneNumbers[1].Id, "Phone 1 wasn't saved.");
        }

        [TestMethod]
        public void LoadPersonWithPhoneNumbers_PhoneNumbersIsNullWithoutExplicitLoad()
        {
            Person original = CreatePersonWithPhoneNumbersAndSave();

            using (MyDbContext context = new MyDbContext())
            {
                var reloaded =
                    (
                    from temp in context.Persons
                    where temp.Id == original.Id
                    select temp
                    ).FirstOrDefault();

                Assert.IsNotNull(reloaded);

                Assert.AreEqual<int>(
                    0, 
                    reloaded.PhoneNumbers.Count, 
                    "Phone number count should be zero.");
            }
        }

        [TestMethod]
        public void LoadPersonWithPhoneNumbers_PhoneNumbersIsNullWithExplicitLoad()
        {
            Person original = CreatePersonWithPhoneNumbersAndSave();

            using (MyDbContext context = new MyDbContext())
            {
                var reloaded =
                    (
                    from temp in context.Persons.Include(x => x.PhoneNumbers)
                    where temp.Id == original.Id
                    select temp
                    ).FirstOrDefault();

                Assert.IsNotNull(reloaded);

                Assert.AreNotEqual<int>(
                    0,
                    reloaded.PhoneNumbers.Count,
                    "Phone number count should not be zero.");

                Assert.AreEqual<int>(original.PhoneNumbers.Count,
                    reloaded.PhoneNumbers.Count, 
                    "PhoneNumber count should match.");
            }
        }

        private Person CreatePersonWithPhoneNumbersAndSave()
        {
            var person = CreatePersonAndSave();

            person.PhoneNumbers.Add(new Phone()
            {
                PhoneDescription = "Work",
                PhoneNumber = "123-456-7890"
            });

            person.PhoneNumbers.Add(new Phone()
            {
                PhoneDescription = "Mobile",
                PhoneNumber = "123-234-3456"
            });

            UpdatePerson(person);
            return person;
        }

        private Person CreatePersonAndSave()
        {
            var person = new Person();

            person.FirstName = UnitTestUtility.GetStringValue("fn");
            person.LastName = UnitTestUtility.GetStringValue("ln");

            using (MyDbContext context = new MyDbContext())
            {
                context.Persons.Add(person);
                context.SaveChanges();
            }

            return person;
        }

        private void LoadPersonModifyAndSave(int id)
        {
            using (MyDbContext context = new MyDbContext())
            {
                var person = context.Persons.Find(id);

                if (person == null)
                {
                    Assert.Fail("Could not find person with id '{0}'.", id);
                }

                person.FirstName = UnitTestUtility.GetStringValue("modifiedfn");
                person.LastName = UnitTestUtility.GetStringValue("modifiedln");

                context.SaveChanges();
            }
        }

        private Person LoadExistingPerson(int id)
        {
            using (MyDbContext context = new MyDbContext())
            {
                var person = context.Persons.Find(id);

                if (person == null)
                {
                    Assert.Fail("Could not find person with id '{0}'.", id);
                }

                return person;
            }
        }

        private void UpdatePerson(Person saveThis)
        {
            using (MyDbContext context = new MyDbContext())
            {
                var entity = context.Attach<Person>(saveThis);

                entity.State = EntityState.Modified;

                context.SaveChanges();
            }
        }
    }
}
